package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** an efficient representation of a collection of trees (Tomita's shared packed parse forest) */
public abstract class Forest<T> {

    /** assume that this forest contains exactly one tree and return it; otherwise throw an exception */
    public final Tree<T> expand1() throws Parser.Ambiguous, Parser.Failed {
        Iterator<Tree<T>> it = expand(true).iterator();
        if (!it.hasNext()) throw new Parser.Failed();
        return it.next();
    }

    /** expand this forest into a set of trees */
    public abstract HashSet<Tree<T>>  expand(boolean toss);

    abstract boolean valid();

    static <T> Forest<T>   singleton(Token.Location loc, Sequence creator)                       { return create(loc, null, new Forest[] { }, creator, false, true); }
    static <T> Forest<T>   singleton(Token.Location loc, Forest<T> body, Sequence creator)       { return create(loc, null, new Forest[] { body }, creator, false, true); }
    static <T> Forest<T>   leaf(Token.Location loc, T tag,                     Sequence creator) { return create(loc, tag, null, creator, false, false); }
    public static <T> Forest<T>   create(Token.Location loc, T tag, Forest<T>[] tokens, Sequence creator, boolean unwrap, boolean singleton) {
        return new MultiForest<T>(loc, tag, tokens, creator, unwrap, singleton);
    }

    // Body //////////////////////////////////////////////////////////////////////////////

    protected static class Body<T> {

        private final Token.Location    location;
        private final T                 tag;
        private final Forest<T>[]       tokens;
        private final Sequence          creator;
        private final boolean           unwrap;
        private final boolean           singleton;

        private Body(Token.Location loc, T tag, Forest<T>[] tokens, Sequence creator, boolean unwrap, boolean singleton) {
            this.location = loc;
            this.tag = tag;
            this.tokens = tokens==null ? emptyForestArray : new Forest[tokens.length];
            if (tokens != null) System.arraycopy(tokens, 0, this.tokens, 0, tokens.length);
            this.creator = creator;
            this.unwrap = unwrap;
            this.singleton = singleton;
        }

        private HashSet<Tree<T>> expand(boolean toss, ArrayList<Tree<T>> toks, int i, HashSet<Tree<T>> h) {
            if (singleton) {
                for(Body<T> b : (IterableForest<T>)tokens[0]) b.expand(toss, toks, i, h);

            } else if (i==tokens.length) {
                h.add(new Tree<T>(null, tag, toks.toArray(tree_hint)));

            } else if (unwrap && i==tokens.length-1) {
                if (tokens[i] != null)
                    for(Body b : (IterableForest<T>)tokens[i])
                        b.expand(toss, toks, 0, h);

            } else {
                if (tokens[i]!=null) {
                    HashSet<Tree<T>> exp = tokens[i].expand(toss);
                    if (exp != null)
                        for(Tree<T> r : exp) {
                            int old = toks.size();
                            toks.add(r);
                            expand(toss, toks, i+1, h);
                            while(toks.size() > old) toks.remove(toks.size()-1);
                        }
                }
            }
            return h;
        }
        void addTo(HashSet<Body> h) {
            if (!singleton) h.add(this);
            else for(Body b : (IterableForest<T>)tokens[0]) b.addTo(h);
        }
        void addTo(FastSet<Body> h) {
            if (!singleton) h.add(this, true);
            else for(Body b : (IterableForest<T>)tokens[0]) b.addTo(h);
        }

        private boolean kcache = false;
        private boolean keep = false;
        public boolean keep() {
            if (kcache) return keep;
            kcache = true;
            for(Forest<T> token : tokens) if (!token.valid()) return keep = false;
            return keep = creator==null || (creator.needs.size()==0 && creator.hates.size()==0);
        }
        public boolean keep(Iterable<Body<T>> h) {
            if (keep()) return true;
            for(Forest<T> token : tokens) if (!token.valid()) return false;
            int needs = 0;
            for(Body<T> b : h) {
                if (creator.hates.contains(b.creator) && b.keep(h)) return false;
                if (creator.needs.contains(b.creator) && b.keep(h)) needs--;
            }
            return needs <= -1 * creator.needs.size();
        }


        public String toString() {
            StringBuffer ret = new StringBuffer();
            for(int i=0; i<tokens.length; i++) {
                String q = tokens[i]==null ? "null" : tokens[i].toString();
                if (q.length() > 0) {
                    ret.append(q);
                    ret.append(" ");
                }
            }
            String tail = ret.toString().trim();
            String head = (tag!=null && !tag.toString().equals("")) ? (tail.length() > 0 ? tag+":" : tag+"") : "";
            if (tail.length() > 0) tail = "{" + tail + "}";
            return head + tail;
        }
    }


    // Ref //////////////////////////////////////////////////////////////////////////////

    static abstract class IterableForest<T> extends Forest<T> implements Iterable<Forest.Body<T>> {
        public abstract Iterator<Forest.Body<T>> iterator();
    }

    /**
     *  This class represents a partially complete collection of
     *  forests to be viewed as a forest at some later date; once
     *  viewed, it becomes immutable
     */
    static class Ref<T> extends IterableForest<T> {
        private FastSet<Forest> hp = new FastSet<Forest>();
        private Forest res = null;
        public boolean valid = false;
        public Ref() { }
        public void merge(Forest p) {
            if (res != null) throw new Error("already resolved!");
            if (p==null) throw new Error();
            if (p!=this) hp.add(p);
        }
        public Iterator<Body<T>> iterator() { return ((IterableForest<T>)resolve()).iterator(); }
        public HashSet<Tree<T>> expand(boolean toss) { return resolve().expand(toss); }
        public boolean valid() { if (valid) return true; resolve(); return valid; }
        public String toString() { return resolve().toString(); }
        public Forest resolve() {
            if (hp==null) return res;
            HashSet<Body> results = null;
            FastSet<Body> nh      = new FastSet<Body>();
            for(Forest<?> p : hp)
                for(Body<?> b : (IterableForest<?>)p) {
                    if (b.keep() && (b.creator==null || !b.creator.lame)) { valid = true; b.addTo(nh); }
                    else results = new HashSet<Body>();
                }
            if (results != null) {
                for(Forest<?> p : hp)
                    for(Body<?> b : (IterableForest<?>)p)
                        results.add(b);
                for(Body b : results) {
                    if (b.keep() && (b.creator==null || !b.creator.lame)) continue;
                    if (b.creator!=null && b.creator.lame) continue;
                    if (!b.keep(results)) continue;
                    valid = true;
                    b.addTo(nh);
                }
            }
            hp = null;
            return res = new MultiForest(nh, valid);
        }
    }

    // Implementations //////////////////////////////////////////////////////////////////////////////

    private static class MultiForest<T> extends IterableForest<T> {
        private final FastSet<Body<T>> results;
        private boolean valid;
        public boolean valid() { return valid; }
        private MultiForest(FastSet<Body<T>> results, boolean valid) { this.results = results; this.valid = valid; }
        public MultiForest(Token.Location loc, T tag, Forest<T>[] tokens, Sequence creator, boolean unwrap, boolean singleton) {
            this.results = new FastSet<Body<T>>(new Body(loc, tag, tokens, creator, unwrap, singleton));
            this.valid = true;
        }
        public Iterator<Body<T>> iterator() { return results.iterator(); }

        public HashSet<Tree<T>> expand(boolean toss) {
            HashSet<Tree<T>> ret = new HashSet<Tree<T>>();
            for(Body<T> b : results)
                ret.addAll(b.expand(toss, new ArrayList<Tree<T>>(), 0, new HashSet<Tree<T>>()));
            if (toss && ret.size() > 1) throw new Parser.Ambiguous(this);
            return ret;
        }
        
        // Display //////////////////////////////////////////////////////////////////////////////
        
        private String toString = null;
        public String toString() {
            if (toString != null) return toString;
            StringBuffer ret = new StringBuffer();
            if (results.size()==1) {
                for(Forest.Body<T> r : results)
                    ret.append(r);
                return toString = ret.toString();
            }
            ret.append("<?");
            boolean first = true;
            for(Forest.Body<T> r : results) {
                if (!first) ret.append(' ');
                first = false;
                ret.append(r);
            }
            ret.append("?>");
            return toString = ret.toString();
        }
    }

    // Statics //////////////////////////////////////////////////////////////////////////////

    private static Tree[] tree_hint = new Tree[0];
    private static Body[] body_hint = new Body[0];
    private static final Forest[] emptyForestArray = new Forest[0];
}
